
package aa.core;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.lang.Class;
import java.lang.reflect.Constructor;

import aa.gui.View;
import aa.tool.SystemThread;
import aa.util.Queue;

/**
 * This class represents an actor platform.
 * 
 *  <p>
 *  <b>History:</b>
 *  <ul>
 * 	<li> February 19, 2004 - version 0.0.2
 * 	<ul>
 * 	    <li> do minor modification.
 * 	</ul>
 *      <li> March 17, 2003 - version 0.0.1
 *      <ul>
 *          <li> create this file.
 *      </ul>
 *  </ul>
 * 
 * @author Myeong-Wuk Jang
 *  @version $Date: 2008/01/01 00:23:49 $ $Revision: 1.1 $
 */

public class Platform extends SystemThread
{
    // ========== Public Constants ========== //
    
    /**
     *  Indicates whether debuging messages are printed on the screen. 
     */
    public final static boolean bDEBUG = false;

    /**
     *  Indicates whether error messages are printed on the screen. 
     */
    public final static boolean bERROR = true;

    /**
     *  Actor communication message
     *  that indicates that a mobile actor is destroyed, and hence, 
     *  information about the agnet should be removed.
     */    					
    public final static String MSG_DEREGISTER_AGENT = "deregisterActor";

    /**
     *  Actor communication message 
     *  that includes a mobile actor.
     */
    public final static String MSG_RECEIVE_AGENT    = "receiveActor";
    
    /**
     *  Actor communication message
     *  that indicates that a mobile actor starts its migration.
     */
    public final static String MSG_TRANSITION_START = "informTransitionStart";
    
    /**
     *  Actor communication message
     *  that indicates that a mobile actor finishes its migration.
     */
    public final static String MSG_TRANSITION_END   = "infromTransitionEnd";

    /**
     *  Actor communicaiton message
     *  that requests an actor creation.
     */
    public final static String MSG_CREATE_ACTOR     = "createActor";


    // ========== Public Variables ========== //
    
    /**
     *  Hasth table to be used for delivering a name of a new actor 
     *  from the actor thread class to the actor behavior class 
     */
    public static Hashtable g_htCreators = new Hashtable();
    

    // ========== Object Contants ========== //

    private static final int ID_PLATFORM = 0;
    					//  identifier of platform.
    private static final int ID_DF = 1;	
    					//  identifier of the default Directory Manager.


    // ========== Object Variables ========== //

    private static Platform m_pPlatform;
    					//  reference to this actor platform.

    private static int m_iID_AGENT;	//  index of a new UAN.
    private static String m_strHostAddress;
    					//  InetAddress of this platform.
    
    private static int m_iID_MSG;	//  index of a new message.


    private int m_iAppActorStartID;	//  start number for identifications of application actors.

    private ActorName m_anPlatform;	//  actor name of this platform.

    private Queue m_qMsgQueue;		//  in-coming message queue of this actor.

    private View m_viewMain;		//  GUI of this platform.

    private ActorManager m_amActorManager;   
    					//  reference to the Actor Manager of this platform.
    private MessageManager m_mmMessageManager;
    					//  reference to the Message Manager of this platform.
    private MigrationManager m_mmMigrationManager;
    					//  reference to the Migration Manager of this platform.

    private ActorName m_anDirectoryManager;
    					//  reference to the default Directory Manager of this platform.
    
    
    
    private CyberOrgManager m_comCyberOrgManager; //reference to the CyberOrg Manager of this platform
    


    // ========== Object Methods ========== //

    /**
     *  Creates a mulit-actor platform.
     * 
     *  @param p_viewMain the reference to View
     *  @param p_intPort the port number of a server socket for in-coming messages
     */
    public Platform(View p_viewMain, Integer p_intPort)
    {
    	//
    	//  call the constructor of the super-class.
    	//
    	super();

	//
	//  print the title of this program.
	//    	
    	System.out.println("% AA (Actor Architecture) Platform:");
//    	System.out.println("% AAA (Adaptive Actor Architecture) Platform:");
    	
    	//
    	// set class varibles by parameters.
	//
    	m_viewMain = p_viewMain;

    	//
    	//  initialize the Internet address.
    	//
    	try {
    	    InetAddress ia = InetAddress.getLocalHost();
    	    Platform.m_strHostAddress = ia.getHostAddress();

	    System.out.println("     - Host Name:    " + ia.getHostName());
	    System.out.println("     - Host Address: " + ia.getHostAddress());
	    System.out.println();
    	} catch (UnknownHostException e) {
    	    if (Platform.bERROR) {
    	    	System.err.println("% Platform.init: " + e);
    	    	e.printStackTrace();
    	    }
    	    
    	    reportFailure(e.toString());
    	    return;
    	}

	//
	//  initialize object variables.
	// 
    	Platform.m_pPlatform = this;
    	Platform.m_iID_AGENT = Platform.ID_PLATFORM;
    	
    	m_anPlatform = createNewActorName();
    	
    	m_qMsgQueue = new Queue();
    	
	//
	//  inform information about this platform to View.
	//
    	if (m_viewMain != null) {
    	    m_viewMain.setPlatform(this);
    	}

	//
	//  create managers.
	//
    	m_amActorManager = new ActorManager(this, m_viewMain);
    	
    	m_mmMigrationManager = new MigrationManager(this, m_amActorManager);
    	m_mmMigrationManager.start();
    	
    	m_mmMessageManager = 
    	    new MessageManager(this, m_amActorManager, m_mmMigrationManager, p_intPort);
    	m_mmMessageManager.start();
        
        m_comCyberOrgManager = new CyberOrgManager(2,800);

    	//
    	//  set the name of this thread with the class name.
    	//
    	String strClassName = getClass().getName();
    	setName(strClassName.substring(strClassName.lastIndexOf('.') + 1));
    }
    
    
    /**
     *  Reports that starting this platform fails.
     *  <br>
     *  This method is called by the <code>Platform</code> constructor, or
     *  the <code>init</code> method, or the Message Manager.
     * 
     *  @param p_strMsg a message to describe the reason of a failure.
     */
    protected void reportFailure(String p_strMsg)
    {
    	System.out.println("% Starting ths platform failed.");
    	System.out.println("%    - because of " + p_strMsg);
    	
    	if (m_viewMain != null) {
    	   m_viewMain.exit();
    	}
    	
  	System.exit(0);
    }
    

    /**
     *  Processes some necessay routines 
     *  before entering the while loop in the run method .
     *  <br>
     *  This is called by the run method of the super class.
     */
    protected void init()
    {
    	System.out.println("% AA platform started.");
//    	System.out.println("% AAA platform started.");
    	System.out.println();
    	
    	if (m_viewMain != null) {
    	    m_viewMain.printOnStatusBar("AA platform started");
//    	    m_viewMain.printOnStatusBar("AAA platform started");
    	}
    	
    	//
    	//  create the default Directory Manager
    	//
    	try {
    	    m_anDirectoryManager = createActor("aa.app.dm.DirectoryManager", new Object[0]);
    	    
    	    m_iAppActorStartID = Platform.m_iID_AGENT;
            
    	} catch (CreateActorException e) {
    	    System.err.println("% Platform.init: " + e);
    	    reportFailure("the Failure of Creating System Actors");
    	}
    }
    
    
    /**
     *  Uninitializes this program.
     *  <br>
     *  This method is called by the 'View' class.
     */
    public void uninit()
    {
    	m_amActorManager.uninit();
    	m_mmMessageManager.uninit();
    	m_mmMigrationManager.uninit();
    	    	
    	System.out.println("% AA Platform ended.");
//    	System.out.println("% AAA Platform ended.");
    }
    

    /**
     *  Processes a communcation message.
     * 
     *  @param p_amMsg the communication message.
     */
    protected void processMessage(ActorMessage p_amMsg)
    {
    	String strMethod = p_amMsg.getMethod();
    	Object[] oaArg = p_amMsg.getArguments();
    	
    	if (p_amMsg.isErrorMessage()) {
    	    if (m_viewMain != null) {
    	    	m_viewMain.displayErrorMessage(p_amMsg.getErrorMessage());
    	    }
    	    
    	    if (strMethod.equalsIgnoreCase(Platform.MSG_RECEIVE_AGENT)) {
     	        m_mmMigrationManager.receiveActor((String)oaArg[0],
    	    					  (byte[])oaArg[1],
    	    					  (byte[])oaArg[2],
    	    				           p_amMsg.getSender());    	    					 
    	    }
    	} else if (strMethod.equalsIgnoreCase(Platform.MSG_DEREGISTER_AGENT)) {
    	    m_amActorManager.deregisterMobileActor((ActorName)oaArg[0]);
    	    				    
    	} else if (strMethod.equalsIgnoreCase(Platform.MSG_RECEIVE_AGENT)) {
    	    m_mmMigrationManager.receiveActor((String)oaArg[0],
    	    				      (byte[])oaArg[1],
    	    				      (byte[])oaArg[2],
    	    				      p_amMsg.getSender());
    	    					 
    	} else if (strMethod.equalsIgnoreCase(Platform.MSG_TRANSITION_START)) {
    	    m_mmMigrationManager.informMigrationStart((ActorName)oaArg[0],
    	    					      (String)oaArg[1]);
    	    
    	} else if (strMethod.equalsIgnoreCase(Platform.MSG_TRANSITION_END)) {
    	    m_mmMigrationManager.informMigrationEnd((ActorName)oaArg[0],
    	    					    (String)oaArg[1]);
    	    
    	} else if (strMethod.equalsIgnoreCase(Platform.MSG_CREATE_ACTOR)) {
    	    createRemoteActor(p_amMsg);
        //xinghui
        } else if (strMethod.equalsIgnoreCase("receiveCyberOrg")){
            m_comCyberOrgManager.remoteMigrateChild((ActorName)oaArg[0]);
        } else if (strMethod.equalsIgnoreCase("resAlloc")){
            m_comCyberOrgManager.resAlloc((ActorName)oaArg[0],((Long)oaArg[1]).longValue());
        //xinghui
    	} else {
    	    if (Platform.bERROR) {
    	    	System.err.println("> Unknown Method: " + strMethod);
                System.err.println("msg:"+p_amMsg + oaArg.length +oaArg[0].getClass()); 
    	    }
    	}
    }
    
    
    /**
     *  Creates a new actor name.
     * 
     *  @return a new actor name.
     */
    private static synchronized ActorName createNewActorName()
    {
        //System.out.println(Platform.m_iID_AGENT);
    	return new ActorName(Platform.m_strHostAddress, Platform.m_iID_AGENT++);
    }
    
    
    /**
     *  Creates a new message ID.
     * 
     *  @return a new message ID.
     */
    protected static synchronized String creteMessageID()
    {
    	return new String("mid://" + Platform.m_strHostAddress + ":" + Platform.m_iID_MSG++);
    }
    
    
    /**
     *  Returns the reference of this platform.
     * 
     *  @return the reference of this platform.
     */
    protected static Platform getPlatform()
    {
    	return Platform.m_pPlatform;
    }


    /**
     *  Returns the start number for identifications of application actors.
     *  <br>
     *  This method is called by the 'View' class.
     *  
     *  @return the start number for identifications of application actors.
     */
    public int getAppActorStartID()
    {
    	return m_iAppActorStartID;
    }
    

    /**
     *  Returns the Inet adress of this platform.
     * 
     *  @return the Inet adress of this platform.
     */
    protected String getHostAddress()
    {
    	return Platform.m_strHostAddress;
    }
    

    /**
     *  Returns the actor name of this platform.
     *
     *  @return the actor name of this platform.
     */
    public ActorName getActorNameOfPlatform()
    {
        return m_anPlatform;
    }


    /**
     *  Returns the actor name of the specified platform.
     *
     *  @param p_strHostName the host address of a platform.
     * 
     *  @return the actor name of this platform or the specified paltform 
     *               if the platform exists; <br>
     *           <code>null</code> otherwise.
     */
    public ActorName getActorNameOfPlatform(String p_strHostName)
    {
        ActorName anPlatform;
        
        try {
            InetAddress ia = InetAddress.getByName(p_strHostName);
            anPlatform = new ActorName(ia.getHostAddress(), Platform.ID_PLATFORM);
        } catch (UnknownHostException e) {
            if (Platform.bERROR) {
                System.err.println("% Actor.getActorNameOfPlatform: " + e);
                e.printStackTrace();
            }
            
            anPlatform = null;
        }
        
        return anPlatform;
    }
    
    
    /**
     *  Returns the name of the default Direcotry Manager actor.
     * 
     *  @return the name of the default Directory Manager actor.
     */
    protected ActorName getDefaultDirectoryManager()
    {
        return m_anDirectoryManager;
    }


    /**
     *  Returns the name of the default Direcotry Manager actor 
     *  of the specified platform.
     * 
     *  @param p_strHostName the host address of a platform.
     * 
     *  @return the name of the default Directory Manager 
     *               of the specified platform
     * 		     if the platform exists; <br>
     * 		<code>null</code> otherwise.
     */
    protected ActorName getDefaultDirectoryManager(String p_strHostName)
    {
        ActorName anDF;

        try {
            InetAddress ia = InetAddress.getByName(p_strHostName);
            anDF = new ActorName(ia.getHostAddress(), Platform.ID_DF);
        } catch (UnknownHostException e) {
            if (Platform.bERROR) {
                System.err.println("% Actor.getDefaultDirectoryManager: " + e);
                e.printStackTrace();
            }
            
            anDF = null;
        }
        
        return anDF;
    }


    /**
     *  Creates a remote actor.
     *  <br>
     *  This method is initiated by an actor on another platform, and
     *  it is called by the <code>processMessage</code> method.
     * 
     *  @param p_amMsg the communication message.
     */
    private void createRemoteActor(ActorMessage p_amMsg)
    {
    	//
    	//  decompose the input message into the name of a new actor and
    	//  its initial arguments .
    	//
    	Object[] objaMsgArgs = p_amMsg.getArguments();
    	Object[] objaArgs    = new Object[objaMsgArgs.length-1];
    	
	String strActorClass = (String)objaMsgArgs[0];
	
	for (int i=1; i<objaMsgArgs.length; i++) {
	    objaArgs[i-1] = objaMsgArgs[i];
	}
	
	//
	//  create a new actor.
	//
	ActorName anNewActor;
	ActorMessage amMsg;
	
	try {
	    anNewActor = createActor(strActorClass, objaArgs);
	    
	    amMsg = p_amMsg.makeReturnMessage(anNewActor);
	} catch (CreateActorException e) {
	    amMsg = p_amMsg.makeErrorMessage(e.getMessage());
	}
	
	//
	//  reply the in-coming message.
	//
	sendMessage(amMsg);
    }
    

    /**
     *  Creates an actor.
     *  <br>
     *  This method is called by either an actor thread or GUI.
     *  
     *  @param p_strActorClass the class of an actor to be created.
     *  @param p_objaArgs arguments of the constructor of the actor.
     * 
     *  @return the name of a new actor.
     * 
     *  @throws CreateActorException it occurs when actor creation fails.
     */
    public ActorName createActor(String p_strActorClass, Object[] p_objaArgs)
    	throws CreateActorException
    {
    	if (Platform.bDEBUG) {
            System.out.println("> Actor Class Name: " + p_strActorClass); 
    	}
     
  	ActorName actorName = createNewActorName();  	
	ActorThread athread = new ActorThread(actorName, true);
	
	try {
	    athread.setBehavior(p_strActorClass, p_objaArgs, null);
	    
	    //
	    //  if the become operator is called in the constructor of this actor, ...
	    //
	    if (athread.m_bRetire == false) { 
    		m_amActorManager.updateActor(athread, ActorManager.ACTIVE);
	    }
	} catch (ActorException e) {
	    m_amActorManager.removeActor(actorName);
	    throw new CreateActorException(e.getMessage());
	}
     
        //
        //  start the actor.
        //
    	String strName = athread.getActorClassName();
    	
        Thread threadActor = new Thread(athread);
        
        // register this thread to the CyberOrgManger(map ActorName to thread) --xinghui
        m_comCyberOrgManager.registerThread(actorName,threadActor);
        
        //System.out.println("&&&&&&&&&&&&&&&&&&&&&&*********************register thread in m_comCyberOrgManger "+actorName +" "+threadActor);
        
        threadActor.setName(strName.substring(strName.lastIndexOf('.')+1));
        threadActor.start();
        
        //suspend this actor --xinghui
        threadActor.suspend();

        return actorName;
    }
    
    /** Create an Actor.
     * It is called by an actor thread.
     * We want to know the creator, in order to put the new actor into the same cyberorg as 
     * the creator is in
     */
    public ActorName createActor(ActorName p_anCreator,String p_strActorClass,Object[] p_objaArgs){
        
        //create new actor
        ActorName actorName = null;
        try{
            actorName = createActor(p_strActorClass,p_objaArgs);
        }catch (CreateActorException e){
            System.out.println("Exception in actor creation: "+e);
        }
        //put new actor to the same cyberorg as the creator,
        //and schedule it
        
        m_comCyberOrgManager.createActor(p_anCreator,actorName);
        
        
        return actorName;
        
    }
    
    
    /**
     *  Change the behavior of an actor.
     *  <br>
     *  This method is called by an actor.
     *  
     *  @param p_athreadCaller the name of an actor thread that called this method.
     *  @param p_baFlags the internal state flag of the caller actor.
     *  @param p_strActorClass the class name of a new actor to be replaced.
     *  @param p_objaArgs arguments of the constructor of the new actor.
     * 
     *  @return the name of a new actor.
     * 
     *  @throws CreateActorException it occurs when actor creation fails.
     */
    protected void becomeActor(ActorThread p_athreadCaller,
    			       Queue p_qMsgQueue,
    			       boolean[] p_baFlags, 
    			       String p_strActorClass, 
    			       Object[] p_objaArgs)
    	throws BecomeActorException
    {
    	if (Platform.bDEBUG) {
            System.out.println("> Actor Class Name: " + p_strActorClass); 
    	}
     
	//
	//  create a new actor with the existing actor name and security key.
	//
	ActorThread athread = new ActorThread(p_athreadCaller.getActorName(), false);

	try {
	    //
	    // set new behavior of the new actor. 
	    //
	    athread.setBehavior(p_strActorClass, p_objaArgs, p_qMsgQueue);
    	    m_amActorManager.updateActor(athread, ActorManager.ACTIVE);
           
            //
            //  start the actor.
            //
    	    String strName = athread.getActorClassName();
    	    
            Thread threadActor = new Thread(athread);
            
            threadActor.setName(strName.substring(strName.lastIndexOf('.')+1));
            threadActor.start();

        //
        //  handle an exception.
        //
        } catch (ActorException e) {
            //
            //  re-register the caller actor.
            //
      	    updateActor(p_athreadCaller, ActorManager.ACTIVE);
      	    
    	    throw new BecomeActorException(e.getMessage());
        } 
    }
    
    
    /**
     *  Sends a message through the Message manager.
     *  <br>
     *  When the platform receives a message from an actor or other components,
     *  it delivers the message to the Message manager
     *  that will deliver the message to the receiver actor.
     * 
     *  @param p_amMsg an actor communication message.
     */
    public void sendMessage(ActorMessage p_amMsg)
    {
	m_mmMessageManager.deliverMessage(p_amMsg);
    }


    /**
     *  Registers an actor thread to the Actor Manager of this platform.
     *  <br>
     *  This method is (should be) called by an actor thread or
     *  the Migration Manager of this platform.
     * 
     *  @param p_athread the reference to an actor thread.
     *  @param p_intState an actor state.
     */
    protected void registerActor(ActorThread p_athread, Integer p_intState)
    {
        m_amActorManager.registerActor(p_athread, p_intState);
    }
    
    
    /**
     *  Updates information about an actor on the Actor Manager.
     *  <br>
     *  This method is (should be) called by an actor thread.
     * 
     *  @param p_athread reference to an actor thread.
     *  @param p_intState actor state.
     */
    protected void updateActor(ActorThread p_athread, Integer p_intState)
    {
        m_amActorManager.updateActor(p_athread, p_intState);
    }
    
    
    /**
     *  Removes the specified actor from the Actor Manager.
     *  <br>
     *  This method is (shoule be) called by an actor thread.
     * 
     *  @param p_anActor Actor Name of the selected actor.
     */
    protected void removeActor(ActorName p_anActor)
    {
        m_amActorManager.removeActor(p_anActor);
        
        m_comCyberOrgManager.removeActor(p_anActor);
    }
    

    /**
     *  Migrates the specified actor to a different host.
     *  <br>
     *  This method is (should be) called by the actor.
     * 
     *  @param p_anActor the name of an actor to be moved.
     *  @param p_iaDestHost the Internet address of the destination host.
     * 
     *  @throws MigrateActorException it occurs when an exception happens 
     * 		during the actor migration.
     */
    public void migrateActor(ActorName p_anActor, InetAddress p_iaDestHost)
    	throws MigrateActorException
    {
    	m_mmMigrationManager.migrateActor(p_anActor, p_iaDestHost);
    }
    
    
    /**
     *  Changes the port number for the Transprot Manager.
     *  <br>
     *  This method is called by the 'View' class.
     *  
     *  @param p_intPort the new port number for transport manager.
     */
    public void changePortNumber(Integer p_intPort)
    {
    	m_mmMessageManager.changePortNumber(p_intPort);
    }
    
    
    /**
     *  Reports a message.
     * 
     *  @param p_amMsg an actor communicaiton message.
     */
    protected void reportMessage(ActorMessage p_amMsg)
    {
    	if (m_viewMain != null) {
    	    m_viewMain.reportMessage(p_amMsg);
    	}
    }
    
    
    
    /** Methods for cyberorgs  */
    
    /** Create root cyberorg, default CyberOrg class: aa.core.CyberOrg*/
    public Object createCyberOrg(Long p_longTicks,Long p_longTicksRate,Long p_longECash,String p_strFacilitatorClass,Object[] p_objaAgs){
      
        CyberOrg root = new CyberOrg(p_longTicks,p_longTicksRate,p_longECash,p_strFacilitatorClass,p_objaAgs);
        registerCyberOrg(root); //put cyberorg into the cyberorgs list of CyberOrgManager
        return root;
    }
    
    
    /** Create root cyberorg, using user defined cyberorg class*/
    public Object createCyberOrg(String p_strCyberOrgClass, Object[] p_objaAgs){
        Class userCyberOrg; 
        CyberOrg newCyber;
        try{
            userCyberOrg = Class.forName(p_strCyberOrgClass);
            if(p_objaAgs.length==0){
                newCyber = (CyberOrg)userCyberOrg.newInstance();
            }
            else{
                Constructor constructor = findConstructor(userCyberOrg,p_objaAgs);
                newCyber = (CyberOrg)constructor.newInstance(p_objaAgs);
                registerCyberOrg(newCyber);
            }
        }catch(Exception e){
            System.out.println("Exception in creating root CyberOrg: "+e);
            return null;
            
        }
        return newCyber;
    }
    
    public void registerCyberOrg(CyberOrg p_coToBeRegistered){
        m_comCyberOrgManager.registerCyberOrg(p_coToBeRegistered);
    }
    
    public void removeCyberOrg(CyberOrg p_coToBeRemoved){
        m_comCyberOrgManager.removeCyberOrg(p_coToBeRemoved);
    }
    
    /**
     *  Finds a proper constructor.
     *  <br>
     *  This method is called by the createCyberOrg method.
     * 
     *  @param p_classActor the class of an actor to be created.
     *  @param p_objaArgs arguments of the constructor of the actor.
     * 
     *  @return constructor of a new actor.
     * 
     *  @throws NoSuchMethodException it occurs when the construct cannot be found.
     */
    private Constructor findConstructor(Class p_classActor, Object[] p_objaArgs)
    	throws NoSuchMethodException
    {
    	//
    	//  grap all public constructors.
    	//
    	Constructor[] caConstructor = p_classActor.getConstructors();
    	if (caConstructor.length == 0) {
    	    throw new NoSuchMethodException("Class '" + p_classActor + "' has no public constructors!");
    	}
    	
    	//
    	//  find a proper constructor and return it.
    	//
    	for (int i=0; i<caConstructor.length; i++) {
    	    Class[] caParameterType = caConstructor[i].getParameterTypes();
    	    
    	    if (p_objaArgs.length != caParameterType.length) {
    	    	continue;
    	    }
    	    
    	    boolean bFound = true;
    	    for (int j=0; j<p_objaArgs.length; j++) {
    	    	if (!caParameterType[j].isInstance(p_objaArgs[j])) {
    	    	    bFound = false;
    	    	    break;
    	    	}
    	    }
    	    
    	    if (bFound) {
    	    	return caConstructor[i];
    	    }
            
         }
        
        
    	
    	//
    	//  if cannot find a prorper constructor, 
    	//  then report an error.
    	//
        throw new NoSuchMethodException("No constructor for '" + p_classActor + 
      				         p_objaArgs + "'.");
    }

    
    
    /** Resource Return to old host by the actors
     * called by cyberorg isolate
     */
    
    public long returnRes(ActorName p_anActor){
        return m_comCyberOrgManager.returnRes(p_anActor);
    }
    
    /** Reallocate resource for specific actor 
     * called by cyberorg isolate
     */
    
    public boolean resAlloc(ActorName p_anToAlloc, long rateAssign){
        return m_comCyberOrgManager.resAlloc(p_anToAlloc,rateAssign);
    }
    
    /** Local migrate
     * it is called when a cyberorg is migrating to a local destination cyberorg
     */
    
    public void localMigrate(CyberOrg p_cybToMigrate,ActorName p_anDestFacilitator,Contract p_conContract){
        m_comCyberOrgManager.localMigrate(p_cybToMigrate,p_anDestFacilitator,p_conContract);
    }
    
    /** Remote migrate
     * it is called when a cyberorg is migrating to a remote destination cyberorg
     */
    public void remoteMigrate(CyberOrg p_cybToMigrate,ActorName p_anDestFacilitator,Contract p_conContract){
        m_comCyberOrgManager.remoteMigrate(p_cybToMigrate, p_anDestFacilitator, p_conContract);
    }
    
    /** Register actor and its host to the CyberOrgManager
     * it is called in remote migration
     */
    public void registerActorCyb(Object p_anActor,CyberOrg p_cybHost){
        m_comCyberOrgManager.registerActorCyb(p_anActor,p_cybHost);
    }
    
    /** This method is called in remote CyberOrg Migration, after migrating an actor,
     * we need to insert it to the destination cyberorg, reallocate resource to it
     * through the local scheduler 
     */
    public void afterMigrateActor(ActorName p_anActor,Thread p_thread){
        m_comCyberOrgManager.afterMigrateActor(p_anActor,p_thread);
    }
    
    /* This method is called after remote migration, in order to set host for the migrated 
     * facilitator
     */
    public CyberOrg lookupHost(Object p_objFacilitator){
        return m_comCyberOrgManager.lookupHost(p_objFacilitator);
    }
    
    public void killActor(ActorName p_anToKill){
        m_amActorManager.killActor(p_anToKill);
    }
    
    /**
     * These two methods are called when a cyberorg runs out of resource, and when it gets more
     * resources and can be scheduled again.
     */
    public void reactiveCyberOrg(CyberOrg p_cybToActive){
        m_comCyberOrgManager.reactiveCyberOrg(p_cybToActive);
    }
    
    public void disactiveCyberOrg(CyberOrg p_cybToDisactive){
        m_comCyberOrgManager.disactiveCyberOrg(p_cybToDisactive);
    }
    
    public CyberOrg findBestDestination(){
        return m_comCyberOrgManager.lookupDestination();
    }
}
